from datetime import datetime
from typing import Optional

from coiote.utils import ApiEndpoint, api_call, ISO_INSTANT_FORMAT, sanitize_request_param
from coiote.v3.model.device_monitoring import MonitoringStatus, MonitoringDataResponse


class DeviceMonitoring(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="deviceMonitoring")

    @api_call(MonitoringStatus)
    def is_monitored(self, device_id: str) -> MonitoringStatus:
        return self.session.get(self.get_url(f"/configuration/device/{device_id}/enabled"))

    @api_call(MonitoringDataResponse)
    def get_data_batch(self, device_id: str, lwm2m_url: str, start_time: datetime, end_time: Optional[datetime],
                       limit=2048) -> MonitoringDataResponse:
        safe_lwm2m_url = sanitize_request_param(lwm2m_url)
        params = {"timeRangeStart": start_time.strftime(ISO_INSTANT_FORMAT), "limit": limit}
        if end_time:
            params["timeRangeEnd"] = end_time.strftime(ISO_INSTANT_FORMAT)
        return self.session.get(
            self.get_url(f"/data/{device_id}/resourceUrl/{safe_lwm2m_url}"),
            params=params
        )



from dataclasses import asdict
from typing import List, Optional

from coiote.utils import ApiEndpoint, api_call, sanitize_request_param
from coiote.v3.model.setting_values import SettingValue


class SettingValues(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="settingValues")

    @api_call()
    def get_all(self, search_criteria: Optional[str]) -> List[SettingValue]:
        return self.session.get(self.get_url(), params={"searchCriteria": search_criteria})

    @api_call()
    def upsert_one(self, setting_value: SettingValue):
        return self.session.put(self.get_url(), json=asdict(setting_value))

    @api_call(SettingValue)
    def get_device_profile(self, device_id: str, with_device_properties: bool = False) -> List[SettingValue]:
        device_id = sanitize_request_param(device_id)
        params = {"withDeviceProperties": with_device_properties}
        return self.session.get(self.get_url(f"/deviceProfile/{device_id}"), params=params)

    @api_call(SettingValue)
    def get_group_profile(self, group_id: str) -> List[SettingValue]:
        group_id = sanitize_request_param(group_id)
        return self.session.get(self.get_url(f"/groupProfile/{group_id}"))



from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional


@dataclass
class ExecutedTaskInfo:
    taskId: str
    deviceId: str


@dataclass
class TaskReport:
    taskId: str
    deviceId: str
    status: str
    startTime: datetime
    finishTime: datetime
    lastUpdateTime: datetime
    summary: Optional[str]
    blocking: bool
    properties: List[dict]


@dataclass
class TaskReportBatch:
    batch: List[TaskReport]
    nextCursor: int


@dataclass
class TasksSummary:
    totalScheduled: int
    inProgress: int
    completed: int
    successes: int
    failures: int
    notCompleted: int
    completionRate: int
    successRate: int
    failureRate: int



from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Dict, List


@dataclass
class FileLocation:
    @staticmethod
    def create_internal(file_name: str, static_content: bool = False):
        return InternalLocation(InternalLocationData(fileName=file_name, staticContent=static_content))

    @staticmethod
    def create_external(file_url: str, username: Optional[str] = None, password: Optional[str] = None):
        return ExternalLocation(ExternalLocationData(fileUrl=file_url, username=username, password=password))


@dataclass
class InternalLocationData:
    fileName: str
    staticContent: bool = False


@dataclass
class ExternalLocationData:
    fileUrl: str
    username: Optional[str] = None
    password: Optional[str] = None


@dataclass
class InternalLocation(FileLocation):
    InternalLocation: InternalLocationData


@dataclass
class ExternalLocation(FileLocation):
    ExternalLocation: ExternalLocationData


class ResourceCategory(str, Enum):
    FIRMWARE = "FIRMWARE"
    SOFTWARE = "SOFTWARE"
    UNKNOWN = "UNKNOWN"
    IMAGE = "IMAGE"


class ResourceExpirationTime(str, Enum):
    ONE_DAY = "ONE_DAY"
    ONE_WEEK = "ONE_WEEK"
    ONE_MONTH = "ONE_MONTH"
    FOREVER = "FOREVER"


class DownloadProtocol(str, Enum):
    HTTP = "HTTP"
    HTTPS = "HTTPS"
    COAP = "COAP"
    COAPS = "COAPS"
    COAP_TCP = "COAP_TCP"
    COAP_TLS = "COAP_TLS"


@dataclass
class DownloadOptions:
    protocol: DownloadProtocol


@dataclass
class ResourceDownloadData:
    address: str


@dataclass
class Resource:
    name: str
    domain: str
    location: FileLocation
    category: ResourceCategory
    expirationTime: ResourceExpirationTime
    visibleForSubtenants: bool = False
    description: Optional[str] = None
    id: Optional[str] = None
    device: Optional[str] = None
    properties: Dict[str, str] = field(default_factory=dict)
    directGroups: List[str] = field(default_factory=list)


@dataclass
class FileData:
    pass


@dataclass
class Base64FileData(FileData):
    data: str



from dataclasses import dataclass, asdict
from datetime import datetime, timedelta
from enum import Enum
from typing import Any

from requests import Session
from tenacity import retry, wait_exponential, stop_after_attempt, retry_if_not_result


@dataclass
class Credentials:
    username: str
    password: str


class AuthType(Enum):
    CREDENTIALS = 1
    RAW_TOKEN = 2


class Authenticator:
    def __init__(
            self,
            url: str,
            session: Session,
            auth: Any
    ):
        self.url = url
        self.session = session
        self.oauth_url = f"{url}/api/auth/oauth_password"
        self.token = None
        self.token_expiration_time = None

        if isinstance(auth, str):
            self.auth_type = AuthType.RAW_TOKEN
            self.token = auth
            self.token_expiration_time = datetime.now() + timedelta(minutes=30)
            self.creds = None
        elif isinstance(auth, Credentials):
            self.auth_type = AuthType.CREDENTIALS
            self.creds = auth
        else:
            raise ValueError(
                "You must provide either raw token or credentials to authenticate in Coiote DM API")

    def _set_headers(self, bearer_token: str):
        self.session.headers.update(
            {"Authorization": f"Bearer {bearer_token}"})

    def authenticate(self):
        if self.auth_type == AuthType.RAW_TOKEN:
            self._set_headers(self.token)
        elif self.auth_type == AuthType.CREDENTIALS and self.should_acquire_token():
            self.acquire_token()

    def should_acquire_token(self) -> bool:
        if self.token_expiration_time is None:
            return True
        else:
            return datetime.now() + timedelta(seconds=60) > self.token_expiration_time

    @retry(retry=retry_if_not_result(lambda x: x), wait=wait_exponential(multiplier=1, min=4, max=10),
           stop=stop_after_attempt(5))
    def acquire_token(self):
        auth_response = self.session.post(
            self.oauth_url, data=asdict(self.creds))
        if auth_response.status_code == 201:
            json = auth_response.json()
            self.token = json['access_token']
            self.token_expiration_time = datetime.now(
            ) + timedelta(seconds=int(json['expires_in']))
            self._set_headers(self.token)
            return True
        elif auth_response.status_code > 500:
            return False
        else:
            raise ValueError("Failed to acquire auth token")



from dataclasses import asdict, replace

from coiote.utils import ApiEndpoint, api_call
from coiote.v3.model.devices import *


class Devices(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="devices")

    @api_call()
    def get_all(self, search_criteria: Optional[str]) -> List[str]:
        return self.session.get(self.get_url(), params={"searchCriteria": search_criteria})

    @api_call(DevicesBatchAddResponse)
    def create_batch(self, devices: List[Device]) -> DevicesBatchAddResponse:
        updated_devices = self._update_endpoint_names(devices)
        return self.session.post(self.get_url("/batch"), json=[asdict(device) for device in updated_devices])

    @staticmethod
    def _update_properties(device_id: Optional[str], properties: Dict[str, str]) -> Dict[str, str]:
        if device_id is None:
            return properties
        else:
            return dict(properties, endpointName=device_id)

    def _update_device(self, device: Device) -> Device:
        return replace(device, id=None, properties=self._update_properties(device.id, device.properties))

    def _update_endpoint_names(self, devices: List[Device]) -> List[Device]:
        return [self._update_device(device) for device in devices]

    @api_call(DevicesFindResponse)
    def get_device_details(self, search_criteria: Optional[str] = None, fields: List[str] = [], limit: int = 100,
                           pageBookmark: Optional[str] = None) -> DevicesFindResponse:
        return self.session.get(self.get_url("/find/details"), params={
            "searchCriteria": search_criteria,
            "fieldSelection": ",".join(fields),
            "limit": limit,
            "pageBookmark": pageBookmark
        })

    @api_call(SingleDeviceAddResponse)
    def create_one(self, device: Device) -> SingleDeviceAddResponse:
        updated = self._update_device(device)
        return self.session.post(self.get_url(), json=asdict(updated))

    @api_call(Device)
    def get_one(self, device_id: str) -> Device:
        return self.session.get(self.get_url(f"/{device_id}"))

    @api_call()
    def update_one(self, device_id, request: DeviceUpdateRequest) -> None:
        return self.session.put(self.get_url(f"/{device_id}"), json=asdict(request))

    @api_call()
    def delete_one(self, device_id) -> None:
        return self.session.delete(self.get_url(f"/{device_id}"))



from dataclasses import dataclass, field
from typing import Optional, List, Dict
from enum import Enum


@dataclass
class DeviceOperation:
    pass


@dataclass
class ReadDefinition:
    key: str


@dataclass
class ReadOperation(DeviceOperation):
    read: ReadDefinition


@dataclass
class WriteDefinition:
    key: str
    value: str


@dataclass
class WriteOperation(DeviceOperation):
    write: WriteDefinition


@dataclass
class ExecuteArg:
    digit: int
    argument: Optional[str] = None


@dataclass
class ExecuteDefinition:
    key: str
    argumentList: List[ExecuteArg] = field(default_factory=list)


@dataclass
class ExecuteOperation(DeviceOperation):
    execute: ExecuteDefinition


@dataclass
class ConfigurationTaskDefinition:
    name: str
    batchRequests: bool = True
    executeImmediately: bool = True
    operations: List[DeviceOperation] = field(default_factory=list)


class TransferMethod(str, Enum):
    Pull = 'Pull'
    Push = 'Push'


class TransferProtocol(str, Enum):
    HTTP = "HTTP"
    HTTPS = "HTTPS"
    COAP = "COAP"
    COAPS = "COAPS"
    COAP_TCP = "COAP_TCP"
    COAP_TLS = "COAP_TLS"


class UpgradeStrategy(str, Enum):
    ObservationTrigger = "ObservationTrigger"
    WithoutObservations = "WithoutObservations"
    ObservationBased = "ObservationBased"
    SendBased = "SendBased"


@dataclass
class TaskParameter:
    name: str
    value: str


@dataclass
class TaskConfig:
    taskName: str
    parameters: List[TaskParameter]
    properties: Dict[str, str]
    isActive: Optional[bool] = False


@dataclass
class Task:
    id: str
    config: TaskConfig



from dataclasses import asdict
from typing import List, Optional

from coiote.utils import ApiEndpoint, StringResult, api_call, sanitize_request_param
from coiote.v3.model.domains import Domain, DomainUpdateRequest


class Domains(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="domains")

    @api_call()
    def get_all(self, search_criteria: Optional[str] = None) -> List[str]:
        return self.session.get(self.get_url(), params={"searchCriteria": search_criteria})

    @api_call(StringResult)
    def add_domain(self, domain: Domain) -> StringResult:
        return self.session.post(self.get_url(), json=asdict(domain))

    @api_call(Domain)
    def get_one(self, domain_id: str) -> Domain:
        safe_domain = sanitize_request_param(domain_id)
        return self.session.get(self.get_url(f"/{safe_domain}"))

    @api_call()
    def update_one(self, domain_id: str, update: DomainUpdateRequest) -> None:
        safe_domain = sanitize_request_param(domain_id)
        return self.session.put(self.get_url(f"/{safe_domain}"), json=asdict(update))

    @api_call()
    def delete_one(self, domain_id: str) -> None:
        safe_domain = sanitize_request_param(domain_id)
        return self.session.delete(self.get_url(f"/{safe_domain}"))



from dataclasses import dataclass
from typing import Optional


@dataclass
class Domain:
    id: str
    description: Optional[str] = None


@dataclass
class DomainUpdateRequest:
    newDescription: Optional[str]



from coiote.v3.data_model import DeviceDataEntry
from coiote.v3.model.tasks import *


class DeviceClient:
    def __init__(self, device_id: str, coiote_client: 'Coiote') -> None:
        self.device_id = device_id
        self.coiote_client = coiote_client

    def get_all_data(self) -> List[DeviceDataEntry]:
        return self.coiote_client.data_model.get_device_datamodel(self.device_id)

    def write_to_resource(self, human_readable_path: str, value: str):
        print(f"Writing to resource {human_readable_path} - {value}")
        return self.coiote_client.tasks.run_device_config_task(
            device_id=self.device_id,
            task_definition=ConfigurationTaskDefinition(
                name="write-device-dc-api",
                operations=[WriteOperation(
                    write=WriteDefinition(key=human_readable_path, value=value))]
            )
        )

    def execute_resource(self, human_readable_path: str, arguments: List[ExecuteArg] = []):
        print(
            f"Executing resource {human_readable_path} with arguments {arguments}")
        return self.coiote_client.tasks.run_device_config_task(
            device_id=self.device_id,
            task_definition=ConfigurationTaskDefinition(
                name="execute-device-dc-api",
                operations=[ExecuteOperation(
                    execute=ExecuteDefinition(key=human_readable_path, argumentList=arguments))]
            )
        )

    def read_resource(self, human_readable_path: str) -> str:
        print(f"Reading resource {human_readable_path}")
        return self.coiote_client.tasks.run_device_config_task(
            device_id=self.device_id,
            task_definition=ConfigurationTaskDefinition(
                name="read-device-dc-api",
                operations=[ReadOperation(
                    read=ReadDefinition(key=human_readable_path))]
            )
        )

    def get_device_data(self) -> List[DeviceDataEntry]:
        return self.coiote_client.data_model.get_device_datamodel(self.device_id)

    def get_resource_value(self, human_readable_path: str) -> Optional[DeviceDataEntry]:
        result = self.coiote_client.data_model.get_device_datamodel(
            self.device_id, parameters=[human_readable_path])
        matching_result = [
            entry for entry in result if entry.name == human_readable_path]
        if len(matching_result) == 1:
            return matching_result[0]
        else:
            return None

    def reboot_device(self):
        return self.coiote_client.tasks.run_device_config_task(
            device_id=self.device_id,
            task_definition=ConfigurationTaskDefinition(
                name="reboot-device-dc-api",
                operations=[ExecuteOperation(
                    execute=ExecuteDefinition(key="Device.0.Reboot"))]
            )
        )

    def get_location(self):
        return self.coiote_client.data_model.get_device_datamodel(self.device_id, parameters=["Location."])

    def get_task_report(self, task_id: str):
        return self.coiote_client.task_reports.get_report_for_device_task(task_id, self.device_id)



from coiote.utils import ApiEndpoint, api_call
from coiote.v3.model.dialects import Dialect


class Dialects(ApiEndpoint):
    def __init__(
            self, *args, **kwargs
    ):
        super().__init__(*args, **kwargs, api_url="dialects")

    @api_call(Dialect)
    def add_lwm2m_object_definition(self, xml_definition: str, overwrite: bool = False) -> Dialect:
        url = self.get_url("/addObject")
        headers = {'Content-Type': 'text/xml'}
        if overwrite:
            return self.session.put(url, data=xml_definition, headers=headers)
        else:
            return self.session.post(url, data=xml_definition, headers=headers)



from dataclasses import dataclass
from datetime import datetime
from dateutil import parser
import functools
from typing import Any, List, Optional, Dict
from dacite import from_dict, config
from requests import Response, Session
from urllib.parse import quote as sanitize_param

from coiote.auth import Authenticator
from coiote.v3.model.devices import parse_psk, DtlsPreSharedKey

TYPE_HOOKS = {
    datetime: parser.isoparse,
    DtlsPreSharedKey: parse_psk
}

ISO_INSTANT_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


@dataclass
class StringResult:
    result: str


class ApiEndpoint:
    def __init__(self,
                 root_url: str,
                 api_url: str,
                 authenticator: Authenticator,
                 session: Session):
        self.root_url = root_url
        self.authenticator = authenticator
        self.session = session
        self.api_url = api_url

    def get_url(self, endpoint: str = "") -> str:
        return f"{self.root_url}/{self.api_url}{endpoint}"


def deserialize_one(json, dataclass: Optional[Any]):
    return from_dict(data_class=dataclass, data=json, config=config.Config(type_hooks=TYPE_HOOKS))


def deserialize(json, dataclass: Optional[Any]):
    if dataclass is None:
        return json
    elif isinstance(json, list):
        return deserialize_list(json, dataclass)
    else:
        return deserialize_one(json, dataclass)


def deserialize_list(json: List[Any], dataclass):
    return [deserialize_one(element, dataclass) for element in json]


def deserialize_dict(json: Dict[Any, Any], key_dataclass, value_dataclass):
    return {deserialize(key, key_dataclass): deserialize(value, value_dataclass) for key, value in json.items()}


def _safe_get_json(response: Optional[Response]):
    try:
        return response.json()
    except:
        return None


def _make_error(msg: str, response: Optional[Response]):
    json = _safe_get_json(response)
    headers = response.headers if response else None
    return ValueError(f"{msg}\nHeaders:{headers}\nBody: {json}")


def handle_response(response: Response, dataclass: Optional[Any]):
    try:
        if response.status_code != 204 and response.ok:
            return deserialize(json=response.json(), dataclass=dataclass)
        elif not response.ok:
            raise _make_error("Received error response from the server", response)
        else:
            return None
    except:
        if response.status_code == 404:
            return None
        elif response.status_code == 403:
            raise _make_error("Lack of required permissions", response)
        elif response.status_code == 401:
            raise _make_error(
                "Unauthorized even though request was authenticated.", response)
        elif 500 > response.status_code >= 400:
            raise _make_error("Invalid request", response)
        elif response.status_code >= 500:
            raise _make_error(
                f"Internal server error - {response.status_code}", response)
        elif response.ok:
            raise _make_error("Failed to parse the server response.", None)


def api_call(response_format: Optional[Any] = None):
    def decorator_deserialize(func):
        @functools.wraps(func)
        def wrap_with_response_handle(self: ApiEndpoint, *args, **kwargs):
            self.authenticator.authenticate()
            response = func(self, *args, **kwargs)
            return handle_response(response, response_format)

        return wrap_with_response_handle

    return decorator_deserialize


def sanitize_request_param(param: Optional[str]) -> Optional[str]:
    if param is None:
        return None
    else:
        return sanitize_param(param, safe='')
